require 'thor'
require 'benchmark'
require 'zlib'
require 'csv'
require 'fileutils'
require 'anamo/api'

module Anamo
  module Fstree

    class Writer

      def initialize path = '.', prefix = nil, base_path = '/'
        @row_count = 0
        @previous_row = nil
        @handle = nil
        @files_per_file = 1000000
        @path = path
        @prefix = prefix
        @base_path = base_path
        init_handle
      end

      def init_handle
        @handle = Zlib::GzipWriter.open("#{@path}/#{@prefix ? "#{@prefix}-" : ""}#{(@row_count/@files_per_file).round}.gz")
        @handle.write "[#{@base_path}]\n"
      end

      def close_handle
        @handle.close if @handle
      end

      def reinit_handle
        close_handle
        GC.start
        init_handle
      end

      def add_row row
        reinit_handle if @row_count > 0 && @row_count % @files_per_file == 0
        [0,2,3,4,5].each { |i| row[i] = '' if @previous_row[i] == row[i] } if @previous_row
        @handle.write row.to_csv
        @row_count = @row_count + 1
        @previous_row = row
      end

      def close
        @handle.close
      end

    end

    class Traverser

      attr_reader :base_path

      def initialize base_path = '/', writer = Writer.new, max_depth = 2, exclusions = []
        @max_depth = max_depth
        @base_path = base_path
        @writer = writer
        @exclusions = exclusions
      end

      def compute
        walk_path @base_path
        @writer.close
      end

      private

      def walk_path path, parent_data = {}, depth = 0

        begin
          return if @exclusions.include? File.realpath(path) # path is in excluded set
        rescue
          return # path does not exist
        end

        begin
          stat = File.stat(path)
        rescue
          stat = nil
        end

        data = [depth, File.basename(path)]

        open_as_directory = false

        if File.symlink? path
          data << 'sym'
        elsif File.directory? path
          open_as_directory = true
          data << 'd'
        elsif File.file? path
          data << 'f'
        elsif File.socket? path
          data << 'soc'
        elsif File.blockdev? path
          data << 'bd'
        elsif File.chardev? path
          data << 'cd'
        end

        begin
          data << stat.uid
        rescue
          data << '-'
        end

        begin
          data << stat.gid
        rescue
          data << '-'
        end

        begin
          data << stat.mode
        rescue
          data << '-'
        end
        
        begin
          data << stat.mtime
        rescue
          data << '-'
        end

        begin
          data << stat.size
        rescue
          data << '-'
        end

        @writer.add_row data

        if open_as_directory
          sub_depth = depth + 1
          return if sub_depth > @max_depth
          Dir["#{path}/*"].each do |sub_path|
            walk_path sub_path, data, sub_depth
          end
        end

      end

    end

    class Thor < ::Thor

      description = "Inspect the file system tree and send the results"
      desc "fstree exec", description
      long_desc description

      def exec

        inspect
        send

      end

      description = "Inspect the file system tree"
      desc "fstree inspect", description
      long_desc description

      def inspect

        FileUtils.rm_rf temp_folder
        FileUtils.mkdir_p temp_folder

        idx = 0

        paths.each do |path, depth|
          traverser = Traverser.new(path, Writer.new(temp_folder, idx, path), depth, exclusions)
          traverser.compute
          idx = idx + 1
        end

      end

      description = "Send the last version of the file system tree from inspect command to the API"
      desc "fstree send", description
      long_desc description

      def send

        files = {}
        Dir["#{temp_folder}/*"].each do |file|
          files[File.basename(file, '.*')] = UploadIO.new(File.new(file), "application/gzip", File.basename(file))
        end

        ::Anamo::Api.new.post_fstree files

      end

      no_commands do

        def set_paths paths
          @paths = paths
        end

        def paths
          if defined? @paths
            @paths
          else
            return {
              '/' => 2
            }
          end
        end

        def set_exclusions paths
          @exclusions = paths
        end

        def exclusions
          if defined? @exclusions
            @exclusions
          else
            return [
              '/proc',
              '/tmp',
              '/dev',
              '/run'
            ]
          end
        end

      end

      private

      def temp_folder
        "/tmp/anamo/fstree"
      end

    end
  end
end
